"""Sphinx directives."""

from pathlib import Path, PurePosixPath

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.images import Figure, Image

from sphinx_thumb_image.lib import ThumbNodeRequest, format_replacement


class ThumbCommon(Image):
    """Common methods for both thumb image/figure subclassed directives."""

    __option_spec = {}
    __option_spec["resize-width"] = lambda arg: directives.nonnegative_int(arg.replace("px", ""))
    __option_spec["resize-height"] = __option_spec["resize-width"]
    __option_spec["no-resize"] = directives.flag
    __option_spec["target-format"] = directives.flag
    __option_spec["no-target-format"] = directives.flag
    __option_spec["no-default-target"] = directives.flag

    def __default_target(self):
        """Apply the thumb_image_default_target config."""
        if "target" in self.options:
            return
        if "no-default-target" in self.options:
            return
        config = self.state.document.settings.env.config
        default_target = config["thumb_image_default_target"]
        if default_target is not None:
            self.options["target"] = default_target

    def __format_target(self):
        """Apply the "target-format" option."""
        if "target" not in self.options:
            return
        if "no-target-format" in self.options:
            return
        config = self.state.document.settings.env.config
        if "target-format" not in self.options and not config["thumb_image_target_format"]:
            return
        # Build substitutions.
        doctree_source = Path(self.state.document["source"])
        env = self.state.document.settings.env
        subdir = PurePosixPath(doctree_source.parent.relative_to(env.srcdir).as_posix())
        substitutions = {
            "raw_path": self.arguments[0],
            "fullsize_path": str(subdir / self.arguments[0]),
        }
        substitutions.update(config["thumb_image_target_format_substitutions"])
        # Format.
        target = self.options["target"]
        for key, value in substitutions.items():
            if callable(value):
                kwargs = {
                    "self": self,
                    "substitutions": substitutions,
                    "target": target,
                    "env": env,
                }
                replacement = value(**kwargs)
            else:
                replacement = value
            target = format_replacement(target, key, replacement)
        if target == self.options["target"]:
            self.state.document.reporter.warning('no subtitutions made by "target-format" in "target"', line=self.lineno)
        else:
            self.options["target"] = target

    def __add_request(self, sphinx_nodes: list[nodes.Element]) -> list[nodes.Element]:
        """Build and add a ThumbRequest to the image node.

        :param sphinx_nodes: List of nodes returned by super().run(), one of which contains an image node to be modified.

        :return: The same node list as the input with an annotated image node.
        """
        config = self.state.document.settings.env.config

        if "no-resize" in self.options:
            request = ThumbNodeRequest(
                no_resize=True,
            )
        elif "resize-width" in self.options or "resize-height" in self.options:
            # Read width/height from directive options first.
            request = ThumbNodeRequest(
                width=self.options.get("resize-width", None),
                height=self.options.get("resize-height", None),
            )
        else:
            # Read width/height from Sphinx config.
            thumb_image_resize_width = config["thumb_image_resize_width"]
            thumb_image_resize_height = config["thumb_image_resize_height"]
            if thumb_image_resize_width is not None or thumb_image_resize_height is not None:
                request = ThumbNodeRequest(
                    width=thumb_image_resize_width,
                    height=thumb_image_resize_height,
                )
            else:
                # User has not provided the width/height.
                raise self.error('Error in %r directive: "resize-width" option is missing.' % self.name)

        # Add request to the node.
        for node in sphinx_nodes:
            for image_node in node.findall(nodes.image):
                image_node[request.KEY] = request

        return sphinx_nodes


class ThumbImage(ThumbCommon):
    """Thumbnail image directive."""

    option_spec = Image.option_spec | ThumbCommon._ThumbCommon__option_spec

    def run(self) -> list[nodes.Element]:
        """Entrypoint."""
        self._ThumbCommon__default_target()
        self._ThumbCommon__format_target()
        return self._ThumbCommon__add_request(super().run())


class ThumbFigure(Figure, ThumbCommon):
    """Thumbnail figure directive."""

    option_spec = Figure.option_spec | ThumbCommon._ThumbCommon__option_spec

    def run(self) -> list[nodes.Element]:
        """Entrypoint."""
        self._ThumbCommon__default_target()
        self._ThumbCommon__format_target()
        return self._ThumbCommon__add_request(super().run())



"""Miscellaneous code."""

import re
from dataclasses import dataclass
from typing import ClassVar, Optional


@dataclass
class ThumbNodeRequest:
    """Request data to be attached to an image node's attribute list.

    Designed for messages to be passed from the image directives to the resizing class.
    """

    width: Optional[int] = None
    height: Optional[int] = None
    no_resize: Optional[bool] = None

    KEY: ClassVar[str] = "thumb-request"


def format_replacement(target: str, key: str, replacement: str) -> str:
    """Apply special formatting to the replacement string.

    :param target: The unformatted string to operate on.
    :param key: Format key (e.g. "key" in "%(key)s").
    :param replacement: The value to replace the key formatting with.

    :return: The formatted string.
    """
    # Simple slices.
    for search, a, b, c in re.findall(rf"(%\({key}:(-?\d+):(-?\d+):(-?\d+)\)s)", target):
        target = target.replace(search, replacement[int(a):int(b):int(c)])
    for search, a, b in re.findall(rf"(%\({key}:(-?\d+):(-?\d+)\)s)", target):
        target = target.replace(search, replacement[int(a):int(b)])
    for search, a in re.findall(rf"(%\({key}:(-?\d+)\)s)", target):
        target = target.replace(search, replacement[int(a)])
    # Three slice permutations.
    for search, a, b in re.findall(rf"(%\({key}::(-?\d+):(-?\d+)\)s)", target):
        target = target.replace(search, replacement[:int(a):int(b)])
    for search, a, b in re.findall(rf"(%\({key}:(-?\d+)::(-?\d+)\)s)", target):
        target = target.replace(search, replacement[int(a)::int(b)])
    for search, a, b in re.findall(rf"(%\({key}:(-?\d+):(-?\d+):\)s)", target):
        target = target.replace(search, replacement[int(a):int(b):])
    for search, a in re.findall(rf"(%\({key}:(-?\d+)::\)s)", target):
        target = target.replace(search, replacement[int(a)::])
    for search, a in re.findall(rf"(%\({key}::(-?\d+):\)s)", target):
        target = target.replace(search, replacement[:int(a):])
    for search, a in re.findall(rf"(%\({key}:::(-?\d+)\)s)", target):
        target = target.replace(search, replacement[::int(a)])
    # Two slice permutations.
    for search, a in re.findall(rf"(%\({key}::(-?\d+)\)s)", target):
        target = target.replace(search, replacement[:int(a)])
    for search, a in re.findall(rf"(%\({key}:(-?\d+):\)s)", target):
        target = target.replace(search, replacement[int(a):])
    # No slice.
    return target.replace(f"%({key})s", replacement)



"""Sphinx Thumb Image.

Resize images in Sphinx documents/pages to thumbnails.

https://sphinx-thumb-image.readthedocs.io
https://github.com/Robpol86/sphinx-thumb-image
https://pypi.org/project/sphinx-thumb-image
"""

from sphinx.application import Sphinx

from sphinx_thumb_image.directives import ThumbFigure, ThumbImage
from sphinx_thumb_image.resize import ThumbImageResize

__author__ = "@Robpol86"
__license__ = "BSD-2-Clause"
__version__ = "0.2.0"


def setup(app: Sphinx) -> dict[str, str]:
    """Register extension components with Sphinx (called by Sphinx during phase 0 [initialization]).

    :param app: Sphinx application object.

    :returns: Extension version.
    """
    app.add_config_value("thumb_image_resize_width", None, "html")
    app.add_config_value("thumb_image_resize_height", None, "html")
    app.add_config_value("thumb_image_target_format", False, "html")
    app.add_config_value("thumb_image_target_format_substitutions", {}, "html")
    app.add_config_value("thumb_image_default_target", None, "html")
    app.add_directive("thumb-image", ThumbImage)
    app.add_directive("thumb-figure", ThumbFigure)
    app.connect("doctree-read", ThumbImageResize.resize_images_in_document, priority=499)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
        "version": __version__,
    }



"""Image resizing module."""

from os.path import relpath
from pathlib import Path
from typing import Optional

import PIL.Image
import PIL.ImageFile
import PIL.ImageSequence
from docutils.nodes import Element, document
from portalocker import LockException, TemporaryFileLock
from sphinx.application import Sphinx
from sphinx.util import logging

from sphinx_thumb_image.lib import ThumbNodeRequest


class ThumbImageResize:
    """Resize images."""

    THUMBS_SUBDIR = "_thumbs"

    @classmethod
    def save_animated(cls, image: PIL.ImageFile.ImageFile, target: Path, target_size: tuple[int, int]):
        """Save all frames in an animated image file to the target file.

        :param image: Opened source image.
        :param target: Path to target file.
        :param target_size: Image width and height to resize to.
        """
        frames = []
        for frame in PIL.ImageSequence.Iterator(image):
            frame_resized = frame.resize(target_size)
            frames.append(frame_resized)
        disposal = 2  # https://github.com/Robpol86/sphinx-thumb-image/issues/43
        frames[0].save(target, format=image.format, save_all=True, append_images=frames[1:], disposal=disposal)

    @classmethod
    def resize(
        cls, source: Path, target_dir: Path, request: ThumbNodeRequest, doctree: document, node: Element
    ) -> Optional[Path]:
        """Resize one image.

        Output image saved with the same relative path as the source image but in the thumbs directory.

        :param source: Path to image file to resize.
        :param target_dir: Path to directory to write resized output image to.
        :param request: Image node's extension request object.
        :param doctree: Current document.
        :param node: Current image node.

        :returns: Path to the output image or None if no resizing is done.
        """
        log = logging.getLogger(__name__)
        log.debug(f"opening {source}")
        with PIL.Image.open(source) as image:
            source_size = image.size
            is_animated = getattr(image, "is_animated", False) and image.n_frames > 1
            # Get target size.
            if is_animated:
                image_copy = image.copy()
                image_copy.thumbnail((request.width or source_size[0], request.height or source_size[1]))
                target_size = image_copy.size
            else:
                image.thumbnail((request.width or source_size[0], request.height or source_size[1]))
                target_size = image.size
            if target_size[0] >= source_size[0]:
                paren = f"{node['uri']} {source_size[0]}x{source_size[1]}"
                message = f"requested thumbnail size is not smaller than source image ({paren})"
                doctree.reporter.warning(message, source=node.source, line=node.line)
                return None
            # Get target file path.
            thumb_file_name = f"{source.stem}.{target_size[0]}x{target_size[1]}{source.suffix}"
            target = target_dir / thumb_file_name
            if target.exists():
                log.debug(f"skipping {source} ({target} exists)")
                return target
            # Write to target file path.
            target.parent.mkdir(exist_ok=True, parents=True)
            lock_file = target.parent / f"{target.name}.lock"
            try:
                with TemporaryFileLock(lock_file, timeout=0):
                    if target.exists():
                        log.debug(f"skipping {source} ({target} exists after lock)")
                        return target
                    log.debug(f"resizing {source} ({source_size[0]}x{source_size[1]}) to {target}")
                    if is_animated:
                        cls.save_animated(image, target, target_size)
                    else:
                        image.save(target, format=image.format)
            except LockException:
                log.debug(f"skipping {source} ({target} exists after race)")
                return target
        return target

    @classmethod
    def resize_images_in_document(cls, app: Sphinx, doctree: document):
        """Resize all images in one Sphinx document.

        Called from the doctree-read event.

        :param app: Sphinx application object.
        :param doctree: Current document.
        """
        thumbs_dir = app.env.doctreedir / cls.THUMBS_SUBDIR
        doctree_source = Path(doctree["source"])
        for node in doctree.findall(lambda n: ThumbNodeRequest.KEY in n):
            request: ThumbNodeRequest = node[ThumbNodeRequest.KEY]
            if request.no_resize:
                continue
            imguri = node["uri"]
            if imguri.startswith("data:"):
                doctree.reporter.warning("embedded images (data:...) are not supported", source=node.source, line=node.line)
                continue
            if imguri.find("://") != -1:
                doctree.reporter.warning("external images are not supported", source=node.source, line=node.line)
                continue
            path_rel, path_abs = app.env.relfn2path(imguri, app.env.docname)
            source = Path(path_abs)
            if not source.is_file():
                continue  # Subclassed Image directive already emits a warning in this case.
            target_dir = thumbs_dir / Path(path_rel).parent
            target = cls.resize(source, target_dir, request, doctree, node)
            if not target:
                continue
            node["uri"] = relpath(target, start=doctree_source.parent)
